import numpy as np
from operator import mul
from math import ceil
from GenZ.unit import Unit
from GenZ.system import System
from GenZ.Models import OpType, CollectiveType
from GenZ.collective_times import get_AR_time, get_A2A_time, get_message_pass_time, get_AG_time

# 4, 5 Regular Logit and Attend
# 9, 10 Beam Merge Logit and attend
op_type_dicts = {0: 'FC', 1: 'CONV2D', 2: 'DWCONV', 3: 'GEMM', 4: 'Logit', 5: 'Attend', 6:'Sync',
                9:'Logit', 10:'Attend', 11:'CONV1D', 12:'Einsum', 13:'Repeat', 14:'EndRepeat',
                15:'Norm', 16:'Avg', 17:'Special_Func'}
class Operator(object):
    def __init__(self, dim, density=(1.0,1.0,1.0)):
        self.dim = [int(x) if isinstance(x, (int, float, np.int32, np.int64)) else x for x in dim]
        self.density_a, self.density_w, self.density_o = density
        self.input_a, self.input_w, self.output = self.get_tensors()
        self.num_ops = self.get_num_ops()
        self.set_mem_pin(*self.get_default_mem_loc())

    def get_default_mem_loc(self):
        return ['off', 'off', 'off']

    def set_mem_pin(self, input_a=None, input_b=None, output=None):
        if input_a is not None:
            self.input_a_loc = input_a
        if input_b is not None:
            self.input_w_loc = input_b
        if output is not None:
            self.output_loc = output

    def set_tensor(self, input_a=None, input_w=None, output=None):
        if input_a is not None:
            self.input_a = input_a
        if input_w is not None:
            self.input_w = input_w
        if output is not None:
            self.output = output

    def get_density_list(self):
        return [self.density_a, self.density_w, self.density_o]

    def get_op_type(self, dim):
        return op_type_dicts[dim[-1]]

    def get_tensors(self):
        pass

    def get_size(self, tensor):
        return np.prod(tensor)

    # Each kind of operation function will have its own num ops, in which using the layer parameters obtained from the
    # .csv file it will give out number of required ops .
    def get_num_ops(self):
        pass

    def get_dimensions(self):
        return self.get_tensors()

    # For each kind of operator, this returns number of required paramters for that layer type. (Refer operators.py )
    def get_effective_dim_len(self):
        pass

    def get_num_data(self):
        return sum(self.get_sz_list())

    def get_effective_num_data(self, system):
        return sum(self.get_sz_list())


    def get_ideal_memory_time(self, system):
        sz_list = self.get_sz_list()
        memory_time_onchip = 0
        memory_time_offchip = 0
        for tensor_sz in sz_list:
            memory_time_onchip += tensor_sz * system.get_bit_multiplier(type='M')/ system.onchip_mem_bw
            memory_time_offchip += tensor_sz * system.get_bit_multiplier(type='M')/ system.offchip_mem_bw
        return  memory_time_offchip, memory_time_onchip


    def get_compute_time(self, system):
        return self.get_effective_num_ops(system) * system.get_bit_multiplier(type='C')/system.op_per_sec


    def get_effective_num_ops(self, system=None):
        return self.get_num_ops()


# The function returns the size of each of the 3 models parameter for each layer, i.e. input, weights and outputs.
    def get_sz_list(self):
        return list(map(self.get_size, [self.input_a, self.input_w, self.output]))

    def get_loc_list(self):
        return [self.input_a_loc, self.input_w_loc, self.output_loc]

    def get_memory_time(self, system):
        sz_list = self.get_sz_list()
        loc_list = self.get_loc_list()
        memory_time = 0
        ## Assume infinite memory
        for tensor_sz, loc in zip(sz_list, loc_list):
            if loc == 'off':
                bw = system.offchip_mem_bw
            elif loc == 'on':
                bw = system.onchip_mem_bw
            else:
                raise ValueError(f'Wrong bw allocation: {loc}.')
            memory_time += tensor_sz * system.get_bit_multiplier(type='M', data='w' if self.get_op_type(self.dim) == 'GEMM' else 'a')/bw
        return memory_time

    def get_communication_time(self, system):
        '''
            Returns the communication time for the operator in seconds.
        '''
        if self.get_op_type(self.dim) != 'Sync':
            return 0
        else:
            data_size = self.communication_data() * system.get_bit_multiplier(type='M', data='a')
            if system.collective_strategy == 'GenZ':
                if self.collective_type == CollectiveType.AllReduce:
                    return get_AR_time(data_size , self.num_collective_nodes, system) / 1000
                elif  self.collective_type == CollectiveType.All2All:
                    return get_A2A_time(data_size , self.num_collective_nodes, system) / 1000
                elif  self.collective_type == CollectiveType.MessagePass:
                    return get_message_pass_time(data_size, system) / 1000
                elif self.collective_type == CollectiveType.AllGather:
                    return get_AG_time(data_size, self.num_collective_nodes, system) / 1000
                else:
                    raise ValueError(f'Unknown collective type: {self.collective_type}.')
            elif system.collective_strategy == 'ASTRA-SIM':
                from .Astra_sim.get_astra_sim_time import get_astrasim_collective_time, get_network_config
                "ALLREDUCE", "ALLTOALL", "ALLGATHER", "REDUCESCATTER"
                collective_convertion = { CollectiveType.AllReduce: 'ALLREDUCE', CollectiveType.All2All: 'ALLTOALL',
                                CollectiveType.AllGather: 'ALLGATHER', CollectiveType.ReduceScatter: 'REDUCESCATTER', 
                                }
                if system.network_config is None:
                    return max(get_astrasim_collective_time(system=system, collective_type=collective_convertion[self.collective_type],
                                                        collective_size=data_size).values())/1e9
                else:
                    if self.collective_type == CollectiveType.MessagePass:
                        parallelism = "PP"
                    elif self.collective_type == CollectiveType.AllReduce:
                        parallelism = "TP"
                    elif self.collective_type == CollectiveType.All2All:
                        parallelism = "EP"
                    elif self.collective_type == CollectiveType.AllGather:
                        parallelism = "TP"
                    else:
                        raise ValueError(f'Unknown parallelism for collective type: {self.collective_type}.')
                    
                    network_config = get_network_config(network_config = system.network_config, 
                                                        parallelism_heirarchy = system.parallelism_heirarchy,
                                                        parallelism = parallelism)
                    if parallelism == "PP":
                            BW = network_config['bandwidth'][0] 
                            lat = network_config['latency'][0]
                            # TODO : There is a bug with astrasim when num_nodes = 2
                            # Using GenZ for now.
                            temp_sys = System(num_nodes=2, topology='FullyConnected', interchip_link_bw=BW, interchip_link_latency=lat)
                            # pipe_time = max(get_astrasim_collective_time(system=temp_sys,
                            #                                     collective_type="ALLTOALL",
                            #                                     collective_size=data_size/2).values())/1e6
                            pipe_time = get_message_pass_time(data_size, temp_sys) / 1000
                            if len(network_config['npus_count']) > 1:   ## PP over more than 1 dimension, we need average time.
                                # Num hops: (dim[0]-1)*dim[1] + dim[1]-1
                                first_dim_time = pipe_time * (network_config['npus_count'][0]-1) * network_config['dimensions-count'][1] 
                                temp_sys = System(num_nodes=2, topology='FullyConnected',
                                                interchip_link_bw=network_config['bandwidth'][1], interchip_link_latency=network_config['latency'][1]) 
                                # second_dim_time = max(get_astrasim_collective_time(system=temp_sys,
                                #                             collective_type="ALLTOALL",
                                #                             collective_size=data_size/2).values())/1e6 * (network_config['npus_count'][1]-1)
                                second_dim_time = (get_message_pass_time(data_size, temp_sys) / 1000) * (network_config['npus_count'][1]-1)
                                return (first_dim_time + second_dim_time)/(network_config['npus_count'][0] * network_config['npus_count'][1] -1)
                            else:
                                return pipe_time
                    else: 
                        return max(get_astrasim_collective_time(collective_type=collective_convertion[self.collective_type],
                                                        collective_size=data_size, network_config=network_config).values())/1e9

    def get_onchip_occupancy(self):
        sz_list = self.get_sz_list()
        loc_list = self.get_loc_list()
        onchip_mem_occupancy = 0
        for tensor_sz, loc in zip(sz_list, loc_list):
            if loc == 'on':
                onchip_mem_occupancy += tensor_sz

        return onchip_mem_occupancy

    def get_model_characterstics(self, system, unit = Unit()):
        num_ops =  self.get_num_ops()
        num_data = self.get_num_data() * system.get_bit_multiplier(type='M')
        op_intensity = num_ops/num_data  if num_data else 0
        input_a_size, input_w_size, output_size = self.get_sz_list()
        ret = {
            'Layer Name': self.name,
            'Op Type': self.get_op_type(self.dim),
            'Dimension': self.get_dimensions(),
            'Op Intensity': op_intensity,
            f'Num ops ({unit.unit_flop})': unit.raw_to_unit(num_ops, type='O'),
            f'Input_a ({unit.unit_mem})': unit.raw_to_unit(input_a_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Input_w ({unit.unit_mem})': unit.raw_to_unit(input_w_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Output ({unit.unit_mem})': unit.raw_to_unit(output_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Total Data ({unit.unit_mem})': unit.raw_to_unit(sum(self.get_sz_list()), type='M')* system.get_bit_multiplier(type='M'),
        }

        return ret

    def get_roofline(self, system, unit):
        ideal_complete_offchip_time, ideal_complete_onchip_time = self.get_ideal_memory_time(system=system)
        # x2 for ops -> MAC has 1 multiplication and 1 Addition hence 2.
        num_ops = self.get_effective_num_ops(system) * 2
        num_data = self.get_effective_num_data(system) * system.get_bit_multiplier(type='M')
        op_intensity = num_ops/num_data if num_data else 0

        compute_time = self.get_compute_time(system=system)

        compute_time /= system.compute_efficiency
        compute_efficiency = system.compute_efficiency

        memory_time = self.get_memory_time(system=system) / system.memory_efficiency

        comm_time = self.get_communication_time(system=system) / system.comm_efficiency

        ## This is special case when there is no calculations
        if compute_time == 0:
            memory_time = 0
        exec_time = max(compute_time, memory_time, comm_time)
        thrpt = num_ops/exec_time if exec_time else 0
        com_to_mem_ratio = compute_time/memory_time if memory_time else 0
        if com_to_mem_ratio == 0:
            boundedness = 'Collective'
        else:
            boundedness = 'Compute' if com_to_mem_ratio > 1 else 'Memory'

        input_a_size, input_w_size, output_size = self.get_sz_list()

        ret = {
            'Layer Name': self.name,
            'Op Type': self.get_op_type(self.dim),
            'Dimension': self.get_dimensions(),
            'Bound': boundedness,
            'C/M ratio': com_to_mem_ratio,
            'Op Intensity': op_intensity,
            f'Latency ({unit.unit_time})': unit.raw_to_unit(exec_time, type='T'),
            f'Cycles': exec_time*system.frequency,
            f'C Effcy': compute_efficiency,
            f'Num ops ({unit.unit_flop})': unit.raw_to_unit(num_ops, type='O'),
            f'Input_a ({unit.unit_mem})': unit.raw_to_unit(input_a_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Input_w ({unit.unit_mem})': unit.raw_to_unit(input_w_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Output ({unit.unit_mem})': unit.raw_to_unit(output_size, type='M')* system.get_bit_multiplier(type='M'),
            f'Total Data ({unit.unit_mem})': unit.raw_to_unit(sum(self.get_sz_list()), type='M')* system.get_bit_multiplier(type='M'),
            f'Throughput ({unit.unit_compute})': unit.raw_to_unit(thrpt, type='C'),
            f'Compute time ({unit.unit_time})': unit.raw_to_unit(compute_time, type='T'),
            f'Memory time ({unit.unit_time})': unit.raw_to_unit(memory_time, type='T'),
            f'Communication time ({unit.unit_time})': unit.raw_to_unit(comm_time, type='T'),
            f'Compute cycle': compute_time*system.frequency,
            f'Memory cycle': memory_time*system.frequency,
            f'Communication cycle': comm_time*system.frequency,
        }

        return ret













import numpy as np
from math import ceil, lcm
import sys
import inspect

class ModelConfig():
    r"""
    This is the configuration class to store the configuration of a [`Model`]. It is used to instantiate an LLM
    model according to the specified arguments, defining the model architecture.
    Args:
        vocab_size (`int`, *optional*, defaults to 32000):
            Vocabulary size of the model. Defines the number of different tokens that can be represented by the
            `inputs_ids` passed when calling [`Model`]
        hidden_size (`int`, *optional*, defaults to 4096):
            Dimension of the hidden representations.
        intermediate_size (`int`, *optional*, defaults to 11008):
            Dimension of the MLP representations.
        num_hidden_layers (`int`, *optional*, defaults to 32):
            Number of hidden layers in the Transformer decoder.
        num_attention_heads (`int`, *optional*, defaults to 32):
            Number of attention heads for each attention layer in the Transformer decoder.
        num_key_value_heads (`int`, *optional*):
            This is the number of key_value heads that should be used to implement Grouped Query Attention. If
            `num_key_value_heads=num_attention_heads`, the model will use Multi Head Attention (MHA), if
            `num_key_value_heads=1 the model will use Multi Query Attention (MQA) otherwise GQA is used. When
            converting a multi-head checkpoint to a GQA checkpoint, each group key and value head should be constructed
            by meanpooling all the original heads within that group. For more details checkout [this
            paper](https://arxiv.org/pdf/2305.13245.pdf). If it is not specified, will default to
            `num_attention_heads`.
        hidden_act (`str` or `function`, *optional*, defaults to `"silu"`):
            The non-linear activation function (function or string) in the decoder.
    """
    def __init__(
        self,
        model = 'dummy',
        vocab_size=32000,
        max_model_len = 128000,
        hidden_size=4096,
        intermediate_size=11008,
        num_ffi = 1,    ## Number of feed forward parallel in the first up projection
        num_encoder_layers=0,
        num_decoder_layers=32,
        num_attention_heads=32,
        head_dim=None,
        num_key_value_heads=None,
        hidden_act="silu",
        sliding_window=4096,
        # MoE specific parameters
        moe_layer_freq = None,
        num_experts = 1,
        expert_top_k = 1,
        # Mamba specific parameters
        mamba_d_state=None,
        mamba_d_conv=None,
        mamba_expand=None,
        mamba_dt_rank="auto",
        mamba_conv_bias=True,
        mamba_proj_bias=False,
        mamba_layer_freq=None,
        **kwargs,
    ):
        self.model = model
        self.vocab_size = vocab_size
        self.max_model_len = max_model_len      ## Maximum length of the model
        self.num_decoder_layers = num_decoder_layers
        self.num_encoder_layers = num_encoder_layers
        
        self.hidden_size = hidden_size
        self.intermediate_size = intermediate_size
        self.num_ffi = num_ffi
        self.hidden_act = hidden_act

        
        # Attention Parameters
        self.num_attention_heads = num_attention_heads
        self.sliding_window = sliding_window
        # for backward compatibility
        if num_key_value_heads is None:
            num_key_value_heads = num_attention_heads
        if head_dim is None:
            head_dim = self.hidden_size // self.num_attention_heads
        self.num_key_value_heads = num_key_value_heads
        self.head_dim = head_dim
        
        # MoE Parameters
        self.is_moe = num_experts > 1
        self.moe_layer_freq = moe_layer_freq    ## If n, than every nth value is moe layer.
        self.num_experts = num_experts
        self.expert_top_k = expert_top_k

        # Mamba Parameters
        self.mamba_layer_freq = mamba_layer_freq
        self.mamba_d_state = mamba_d_state
        self.mamba_d_conv = mamba_d_conv
        self.mamba_expand = mamba_expand if mamba_expand is not None else 1
        self.mamba_dt_rank = ceil(self.hidden_size / 16) if mamba_dt_rank == "auto" else mamba_dt_rank
        self.mamba_conv_bias = mamba_conv_bias
        self.mamba_proj_bias = mamba_proj_bias
        self.is_mamba = (mamba_d_state is not None) and (mamba_layer_freq is not None)
        


        # Create the 2D list of parameters
        self.layer_type = []
        if self.is_mamba and self.is_moe:
            unique_layers = lcm(self.mamba_layer_freq, self.moe_layer_freq) 
        elif self.is_mamba:
            unique_layers = self.mamba_layer_freq
        elif self.is_moe:
            unique_layers = self.moe_layer_freq
        else:
            unique_layers = 1
        num_repeats = self.num_decoder_layers / unique_layers
        assert num_repeats.is_integer(), "Number of decoder layers must be divisible by the unique layers"
        for i in range(unique_layers):
            # Determine the attention type
            if self.is_mamba and (i % self.mamba_layer_freq == 0):
                attention_type = "Mamba"
            else:
                attention_type = "MHA-global"
        
            if self.is_moe and self.moe_layer_freq and (i % self.moe_layer_freq == 0):
                layer_type = "MoE"
            else:
                layer_type = "Dense"
            
            self.layer_type.append([attention_type, layer_type])
        super().__init__()

    def __str__(self):
        return str(vars(self))

def get_all_model_configs(file_name):
    current_module = sys.modules[file_name]
    model_configs = {}
    for name, obj in inspect.getmembers(current_module):
        if isinstance(obj, ModelConfig):
            model_configs[obj.model] = obj
    return model_configs

class ModelCollection():
    def __init__(self, models=None):
        if models is not None:
            self.models = models
        else:
            self.models = {}

    def add_model(self, model_config):
        if not isinstance(model_config, ModelConfig):
            raise TypeError("model_config must be an instance of ModelConfig")
        self.models[model_config.model] = model_config

    def get_model(self, model_name):
        model_name_lower = model_name.lower()
        for name in self.models:
            if name.lower() == model_name_lower:
                return self.models[name]
        return None

    def remove_model(self, model_name):
        if model_name in self.models:
            del self.models[model_name]

    def list_models(self):
        unique_models = {}
        for key, value in self.models.items():
            if value not in unique_models.values():
                unique_models[key] = value
            else:
                existing_key = next(k for k, v in unique_models.items() if v == value)
                if len(key) > len(existing_key):
                    del unique_models[existing_key]
                    unique_models[key] = value
        return list(unique_models.keys())

    def add_model_collection(self, model_collection):
        if not isinstance(model_collection, ModelCollection):
            raise TypeError("model_collection must be an instance of ModelCollection")
        self.models.update(model_collection.models)

    def __str__(self):
        return str({name: str(config) for name, config in self.models.items()})

from .Model_sets.alibaba import alibaba_models
from .Model_sets.google import google_models
from .Model_sets.microsoft import microsoft_models
from .Model_sets.misc import misc_models
from .Model_sets.mistral import mistral_models
from .Model_sets.meta import meta_models
from .Model_sets.nvidia import nvidia_models

MODEL_DICT = ModelCollection()
MODEL_DICT.add_model_collection(ModelCollection(alibaba_models))
MODEL_DICT.add_model_collection(ModelCollection(google_models))
MODEL_DICT.add_model_collection(ModelCollection(microsoft_models))
MODEL_DICT.add_model_collection(ModelCollection(misc_models))
MODEL_DICT.add_model_collection(ModelCollection(mistral_models))
MODEL_DICT.add_model_collection(ModelCollection(meta_models))
MODEL_DICT.add_model_collection(ModelCollection(nvidia_models))


from .utils import ModdelingOutput, get_inference_system, get_offload_system
from .llm_decode import decode_moddeling
from .llm_prefill import prefill_moddeling
from .best_parallelization import factors, get_various_parallization, get_best_parallization_strategy, get_pareto_optimal_performance
from .platform_size import get_minimum_system_size


from ..default_models import ModelConfig, get_all_model_configs

##### Mistral Models ########
# https://huggingface.co/mistralai/Mistral-7B-Instruct-v0.2/blob/main/config.json
mistral_7b_config = ModelConfig(model='mistralai/Mistral-7B',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=14336, num_decoder_layers=32,
    vocab_size=32000, max_model_len=32*1024, hidden_act="silu",
)
# https://huggingface.co/mistralai/Mixtral-8x7B-Instruct-v0.1/blob/main/config.json
mixtral_8x7b_config = ModelConfig(model='mistralai/Mixtral-8x7B',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=14336, num_decoder_layers=32,
    expert_top_k=2, num_experts=8, moe_layer_freq=1,
    vocab_size=32000, max_model_len=32*1024, hidden_act="silu",
)


# https://huggingface.co/mistralai/Ministral-8B-Instruct-2410/blob/main/config.json
ministral_8b_config = ModelConfig(model='mistralai/Ministral-8B-Instruct-2410',
    hidden_size=4096, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=12288, num_decoder_layers=36,
    vocab_size=131072, max_model_len=32*1024, hidden_act="silu",
    sliding_window=32*1024
)

# https://huggingface.co/mistralai/Mistral-Nemo-Base-2407/blob/main/config.json
mistral_nemo_12b_config = ModelConfig(model='mistralai/Mistral-NeMo-12B-Instruct',
    hidden_size=5120, num_attention_heads=32,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=14336, num_decoder_layers=40,
    vocab_size=131072, max_model_len=1024000, hidden_act="silu",
)

# https://huggingface.co/mistralai/Mixtral-8x22B-Instruct-v0.1/blob/main/config.json
mixtral_8x22b_config = ModelConfig(model='mistralai/Mixtral-8x22B',
    hidden_size=6144, num_attention_heads=48,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=16384, num_decoder_layers=56,
    expert_top_k=2, num_experts=8, moe_layer_freq=1,
    vocab_size=32768, max_model_len=64*1024, hidden_act="silu",
)

# https://huggingface.co/mistralai/Mistral-Small-Instruct-2409/blob/main/config.json
mistral_small_config = ModelConfig(model='mistralai/Mistral-Small',
    hidden_size=6144, num_attention_heads=48,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=16384, num_decoder_layers=56,
    vocab_size=32768, max_model_len=32*1024, hidden_act="silu",
)

# https://huggingface.co/mistralai/Mistral-Large-Instruct-2407/blob/main/config.json
mistral_large_config = ModelConfig(model='mistralai/Mistral-Large',
    hidden_size=12288, num_attention_heads=96,
    num_key_value_heads=8, num_ffi = 2,
    intermediate_size=28672, num_decoder_layers=88,
    vocab_size=32768, max_model_len=128*1024, hidden_act="silu",
)

mistral_models = get_all_model_configs(__name__)
mistral_models.update(
    {
    'mistral_7b': mistral_7b_config,
    'mistralai/Ministral-8B': ministral_8b_config,
    'mistralai/mistral-7b': mistral_7b_config,
    'mistralai/mixtral-8x7b': mixtral_8x7b_config,
    'mixtral_8x7b': mixtral_8x7b_config,
    })


import argparse
import copy
import importlib
import os

try:
    # before refactory folders of chakra
    importlib.import_module("chakra.et_def.et_def_pb2")
    from chakra.et_def.et_def_pb2 import GlobalMetadata
    from chakra.et_def.et_def_pb2 import Node
    from chakra.et_def.et_def_pb2 import NodeType
    from chakra.et_def.et_def_pb2 import AttributeProto as ChakraAttr
    from chakra.third_party.utils.protolib import encodeMessage as encode_message
    from chakra.third_party.utils.protolib import decodeMessage as decode_message
    from chakra.third_party.utils.protolib import openFileRd as open_file_rd
except ImportError:
    # after refactory
    from chakra.schema.protobuf.et_def_pb2 import GlobalMetadata
    from chakra.schema.protobuf.et_def_pb2 import Node
    from chakra.schema.protobuf.et_def_pb2 import NodeType
    from chakra.schema.protobuf.et_def_pb2 import AttributeProto as ChakraAttr
    from chakra.src.third_party.utils.protolib import encodeMessage as encode_message
    from chakra.src.third_party.utils.protolib import decodeMessage as decode_message
    from chakra.src.third_party.utils.protolib import openFileRd as open_file_rd


required_type = {
    "num_ops": "int64_val",
    "tensor_size": "uint64_val",
    "comm_priority": "int32_val",
    "comm_size": "int64_val",
    "comm_src": "int32_val",
    "comm_dst": "int32_val",
    "comm_tag": "int32_val"
}


def process_attr(attr):
    name = attr.name
    if name in required_type:
        new_type = required_type[name]
    else:
        return attr
    ret_attr = ChakraAttr(name=attr.name)
    original_type = attr.WhichOneof("value")
    value = getattr(attr, original_type)
    setattr(ret_attr, new_type, value)
    return ret_attr
    

def process_chakra_node(node):
    attrs = copy.deepcopy(node.attr)
    for _ in range(len(node.attr)):
        node.attr.pop()
    assert len(node.attr) == 0
    for attr in attrs:
        new_attr = process_attr(attr)
        node.attr.append(new_attr)
    if node.type in {NodeType.COMM_COLL_NODE, NodeType.COMM_SEND_NODE, NodeType.COMM_RECV_NODE}:
        node.attr.append(
            ChakraAttr(name="is_cpu_op", bool_val=False)
        )
    return node


def convert_chakra_file(input_filename, output_filename):
    fr = open_file_rd(input_filename)
    fw = open(output_filename, "wb")
    gm = GlobalMetadata()
    node = Node()
    decode_message(fr, gm)
    encode_message(fw, gm)
    while decode_message(fr, node):
        node = process_chakra_node(node)
        encode_message(fw, node)
    fr.close()
    fw.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="fix type issue of chakra attrs, either specify a single file or a folder contains lots of files."
    )
    parser.add_argument(
        "--input_filename",
        type=str,
        default=None,
        required=False,
        help="input chakra filename"
    )
    parser.add_argument(
        "--output_filename",
        type=str,
        default=None,
        required=False,
        help="output chakra filename"
    )
    parser.add_argument(
        "--input_folder",
        type=str,
        default=None,
        required=False,
        help="folder contains input chakra files"
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        default=None,
        required=False,
        help="folder contains input chakra files"
    )
    args = parser.parse_args()
    
    folder_mod = (args.input_folder is not None)
    if folder_mod:
        if (args.input_folder is None) or (args.output_folder is None):
            parser.print_help()
            exit(1)
        os.makedirs(args.output_folder, exist_ok=True)
        files = os.listdir(args.input_folder)
        for file in files:
            if not file.endswith(".et"):
                continue
            input_file = os.path.join(args.input_folder, file)
            output_file = os.path.join(args.output_folder, file)
            convert_chakra_file(input_file, output_file)
    else:
        if (args.input_filename is None) or (args.output_filename is None):
            parser.print_help()
            exit(1)
        convert_chakra_file(args.input_filename, args.output_filename)
    


from collections import OrderedDict
from typing import Optional
from GenZ.unit import Unit
import warnings
from GenZ.system import System

from Systems.system_configs import system_configs

offload_bw = 128

unit = Unit()

class RuntimeBreakdown():
    def __init__(self):
        self.Embedding: float = 0
        self.MHA: float = 0
        self.FFN: float = 0
        self.Collective: float = 0
        self.LA_layers: float = 0
        self.QKVO_layers: float = 0
        self.softmax: float = 0
        self.AR_time: float = 0
        self.A2A_time: float = 0
        self.Send_Recv_time: float = 0

    def __repr__(self):
        variables = vars(self)
        return ', '.join(f'{name}: {value}' for name, value in variables.items())



class ModdelingOutput(dict):
    Latency: float = 0
    Throughput: float = 0
    Runtime_breakdown: Optional[RuntimeBreakdown] = None
    is_offload: Optional[bool] = False

def get_offload_system(system, total_memory_req, debug):
    """Create a new system with offloaded memory connections

    Args:
        system (System): System Definition without offload
        total_memory_req (float): Memory required for the model weights and activations in MB
        debug (bool): Debug flag for printing

    Returns:
        System: System Definition with offload
    """
    total_device_memory = unit.raw_to_unit(system.off_chip_mem_size, type='M')/1024 ## GB
    total_memory_req = total_memory_req/1024 ## GB

    memory_offloaded = total_memory_req - total_device_memory

    if debug:
        print(f'Total Memory Req:{total_memory_req}, Total device mem:{total_device_memory}, Mem Offload:{memory_offloaded}')

    ###############################################################################################
    ### Memory Time = max(min(size_required,size_hbm)/BW_hbm, (size_required-size_hbm)/BW_offload)
    ###############################################################################################

    if memory_offloaded > 0:
        new_offchip_BW = (total_memory_req ) / max(min(total_memory_req,total_device_memory)/unit.raw_to_unit(system.offchip_mem_bw, type='BW'), memory_offloaded/offload_bw) 
        system.set_offchip_mem_bw(new_offchip_BW)
        if debug:
            print(f'New BW:{new_offchip_BW}')
    return system

def get_inference_system(system_name='A100_40GB_GPU', bits='bf16', ceff=1, meff=1, **kwargs):
    ##################################################################################################
    ### System Declaration
    ##################################################################################################
    if isinstance(system_name, str):
        if system_name in system_configs:
            system_name = system_configs[system_name]
        else:
            raise ValueError(f'System mentioned:{system_name} not present in predefined systems. Please use systems from Systems/system_configs')
    if isinstance(system_name, dict):
        if system_name.get('real_values',False) == True:
            NUM_FLOPS = system_name.get('Flops', 320)
            OFFCHIP_MEM_BW = system_name.get('Memory_BW',40)
            per_chip_memory = system_name.get('Memory_size',2000)
            C2C_BW = system_name.get('ICN',150)
            C2C_LL = system_name.get('ICN_LL',2)
    elif isinstance(system_name, System):
        return system_name
    else:
        raise TypeError('System should be weight str or dict with Flops,Memory, ICN values')

    return System(unit,frequency=1000 , flops=NUM_FLOPS, off_chip_mem_size=(per_chip_memory*1024), compute_efficiency=ceff, memory_efficiency=meff,
                    offchip_mem_bw=OFFCHIP_MEM_BW, bits=bits, external_mem_bw=offload_bw, interchip_link_bw=C2C_BW, interchip_link_latency=C2C_LL, **kwargs)